'''

 Plug-in Title : "TexturizeByAreaByColourSE"
 Gimp Menu     : "<Image>/Diego/Quantize" 
 Release       :  1.0    (initial, published in a thread of Gimp-chat)
                  1.1    ADDED WHITE BALANCE PREPROCESS (optional)
                         ADDED CURVE PREPROCESS (optional)
                         CHANGED SOME DEFAULT LAYER MODES
                         ALLOWED CREATION OF AN OUTCOME WITH SAME DIMENSIONS AS ORIGINAL
                         CHANGED DEFAULT FILTERING SUBSTRING FOR PATTERNS/GRADIENTS NAMES TO "None" (was "MyWood")
                         ADDED THE POSSIBILITY TO HAVE AREAS SEPARATOR LOOK LIKE NEEDLEPOINTS
                         SUPPRESSED THE SEPARATOR WHEN AREA IS TOO SMALL
                         (chisel/carve remains but only with extreme roundness values: 0.1 or 1.0)
                  1.1.1  fixed three bugs
                         ADDED THE POSSIBILITY TO HAVE NO AREAS SEPARATORS
                  1.2    REMOVED ALL CALLS TO G'MIC FILTERS and changed the name adding a suffix SE (Special Edition)      
 Author        :  Diego Nassetti (dinasset@gimpchat.com)
 Licence       :  GNU General Public License
 
 System Environment used in development 
 --------------------------------------                
 Platform      : Windows 10 64-bits
 Graphic Apps  : Gimp 2.8.18
 Language      : python as included in Gimp
                 
 Extrn Filters : (Gimp) script_fu_chisel (by Rob Antonishen)
                        available under Gimp-chat Scripts&Plug-ins as "Chisel Or Carve"
                 (Gimp) python_fu_WhiteBalanceStretch (by myself)
                        available under Gimp-chat Scripts&Plug-ins as "White Balance Stretch"
                 (Gimp) python_fu_Stretch_RGB_colours (by myself)
                        available under Gimp-chat Scripts&Plug-ins as "Stretch RGB Colors"
 Std filters   : (Gimp) plug_in_sel_gauss
                 (Gimp) plug_in_unsharp_mask 
                 (Gimp) plug_in_spread  
                 (Gimp) plug_in_neon
                 (Gimp) plug_in_vinvert
                 (Gimp) plug_in_erode
                 (Gimp) plug_in_gimpressionist (2 special presets added: pencildraw / squiggly)
                 (Gimp) plug_in_edge
                 (Gimp) plug_in_colortoalpha
                 (Gimp) plug_in_glasstile
                 
 Addtl Resources:(Gimpressionist) Presets: pencildraw
                 (Gimpressionist) Presets: squiggly
                 (Gimp) Pattern: "Leather-05.jpg"
                 
 Note: this Special Edition is "GMIC-free"
 
 Summary
 -------
 Filter to implement the creation on a NEW IMAGE (dimension is a user option)
 starting from an RGB* input image
 where colours are quantized using the temporary conversion of the image
 from RGB to Indexed (nr of colours is a user option)
 each color area is filled with a SELECTED SET OF TEXTURES (user option).
 Filter has been created on the basis of Gimp-chat thread "Let's carve some wood".
 Even if the main use will be with patterns in wood, it has been generalized
 so that you can use any set of textures you like, patterns or gradients
 (for this reason "wood" does not appear in the filter's name)
 
 Details       
 -------
 Flow implemented to get the final result:      
 
 1. calculate the new image width and height scale for the output (user choice for dimensions)
 2. save and set-up context values
 3. create the new image by appropriately scaling the input image dimensions
 4. create the BG layer for the new image by copying the input layer, scaled;
    add alpha to the layer is not present
 5. if WhiteBalance preprocess selected, apply Gimp external filter WhiteBalanceStretch
    if Curves preprocess selected, apply
    - either a brightening curve (option 1)
    - or a darkening curve (option 2)
 6. create a new layer (dup of the BG) for smoothed colours, SOFTLIGHT
 7. apply G'MIC Anisotropic Smoothing
 8. create a new layer (dup of the Smoothed Layer) for the quantization 
 9. transform the image from RGBA to Indexed, with the user chosen number of colours
 10.create a temporary palette
 11.back to RGBA
 12.load the complete set of textures chosen: array of names and calculated luminosity values
    the textures may be patterns as well as gradients
 13.evaluate the separator option chosen:
    - if Needlepoint:
        - create a specific layer (duplicating the active one)
        - apply it on the quantized layer as a whole (before texturizing)
 14.loop on each different color area, selected by using the palette
    a- the user choice may be for having
       - the best texture matching by luma 
       - the best texture matching by colour
       - the best texture matching by luma and colour (default)
    b- if the colour area (in pixels) is small then the width of the effect is small and the effect is chisel
       if the colour area (in pixels) is big then the width of the effect is big and the effect is carve
       otherwise the width of the effect is medium and the effect is chisel
    c- fill each area with the chosen pattern/gradient (i.e. "texturize it")
       evaluate the separator option chosen:
       - if not Needlepoint:
           - set ccRoundness = 0.1 or 1.0 according to the selection made
           - apply the chisel/carve filter
    d- if Needlepoint:
       - merge the needlepoint layer in "Lighten Only Mode" with the quantized/texturized layer (not chiseled/carved)
    e- if None: just do nothing   
    f- merge the quantized layer into the source layer
 15.set the texturized layer to GRAIN-MERGE
 16.create a new layer from visible and set to SOFT-LIGHT at 45% opacity
 17.if the user chose to perform an artistic post-processing,
    - apply the selected one
      the following artistic post-processes are selectable:
        "Contrast Boost",
        "Graphic Evidence",
        "Anguish",
        "Colored Pencil",
        "Squiggly",
        "Edges",
        "Dithering",
        "Colored Texture".
    if the user chose to perform a colour postprocessing,
    - apply the selected one;
      the following colour post-processes are selectable:
        "Chroma Boost",
        "Color Balance",
        "Sepia",
        "Metallic Look - Silver",
        "Metallic Look - Copper",
        "Metallic Look - BlueSteel",
        "Textured Glass".
 18.if the user chose to move the background layer on top,
    raise BG to the top (mode=normal) and duplicate it to allow
    the use of combined modes and opacities
 19.Analyze the Final Layer luminosity and set the top layer mode to:
    - screen if the image is very dark
    - multiply if the image is very light
    - hard-light otherwise   
 20.flatten the image is requested, recalling nr of colours and type of filler in the drawable name
 
 
 
 ==================================================================================================
 
 ----------------------------------------------------------------
 COPYRIGHT NOTICE
 ----------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 -------------------------------------------------------------------
 other info
 -------------------------------------------------------------------
 
'''
#!/usr/bin/env python
from gimpfu import *
import decimal
import random


# service routines


def find_predominant_color (R, G, B):
    
    R16=int(R/16)
    G16=int(G/16)
    B16=int(B/16)
    maxRGB=max(R16,G16,B16)
    minRGB=min(R16,G16,B16)
    limit = 0.25
    # normalize values if differences are low
    if (maxRGB-minRGB) < int(maxRGB*limit):  # if difference lower than limit of the maximum, decide it's GRAY
        search = "WHT"
    elif (maxRGB==R16):
        if ((R16-G16) >= int(R16*limit)) & ((R16-B16) >= int(R16*limit)): 
            search="RED" 
        elif (R16-B16) >= int(R16*limit):
            search="YEL"
        else:
            search="MAG"
    elif (maxRGB==G16):
        if ((G16-R16) >= int(G16*limit)) & ((G16-B16) >= int(G16*limit)): 
            search="GRE"
        elif (G16-B16) >= int(G16*limit):
            search="YEL"
        else:
            search="TRQ"
    elif (maxRGB==B16):
        if ((B16-R16) >= int(B16*limit)) & ((B16-G16) >= int(B16*limit)): 
            search="BLU"
        elif (B16-G16) >= int(B16*limit):
            search="MAG"
        else:
            search="TRQ"
    else:   # placeholder as default impossible case
        search="WHT"
    #info_msg("the predominant colour of this area is "+search)
    return(search)

    
def choose_appropriate_texture_by_LUMACOLOUR (areaColor, LCOLR, LLUMA, LTEXT):
            
    LLUMAS=[]   # subset of Luma
    LTEXTS=[]   # subset of Names
    ss=0        # default
    #info_msg("areaColor = "+str(areaColor))
    (R, G, B, T)=areaColor
    search = find_predominant_color (R, G, B)
    #info_msg("the predominant colour of this area is "+search)
    il=len(LCOLR)
    for i in range(0, il):
        if search==LCOLR[i]:
            chosenColr=LCOLR[i] 
            chosenLuma=LLUMA[i]
            chosenName=LTEXT[i]
            LLUMAS.append(chosenLuma)
            LTEXTS.append(chosenName)
            ss+=1
    
    if ss == 0:    # it may happen that there is no similar colour
        # use random texture
        rn = random.randrange (0, len(LTEXT))
        textureName=LTEXT[rn]
        info_msg ("warning: no appropriate colour found, chosen texture random = "+textureName)
    else:    
        textureName=choose_appropriate_texture_by_LUMA(areaColor, LLUMAS, LTEXTS)
    return(textureName)

def choose_appropriate_texture_by_LUMA (areaColor, LLUMA, LTEXT):
            
    closest="**NONE**"              # default
    distance=999999                 # limit
    (R, G, B, T)=areaColor 
    areaf=int(float(R*0.299)+float(G*0.587)+float(B*0.114))  #new, based on luminosity
    ##info_msg("this area color formula is "+str(areaf))
    il=len(LLUMA)
    for i in range(0, il):
        this_distance=abs(areaf-LLUMA[i])
        ##info_msg("the distance from "+LP[i]+" is "+str(this_distance))
        if this_distance<distance:
            distance=this_distance
            closest=LTEXT[i]
    textureName=closest
    #info_msg("closest = "+textureName)
    
    return(textureName)

def choose_appropriate_texture_by_COLOUR (areaColor, LCOLR, LTEXT):
            
    chosen="**NONE**"              # default
    #info_msg("areaColor = "+str(areaColor))
    (R, G, B, T)=areaColor
    search = find_predominant_color (R, G, B)
    #info_msg("the predominant colour of this area is "+search)
    il=len(LCOLR)
    for i in range(0, il):
        if search==LCOLR[i]:
            chosen=LCOLR[i] 
            textureName=LTEXT[i]
            #info_msg("chosen = "+textureName)
            break
    
    if chosen == "**NONE**":    # it may happen that there is no similar colour
        # use random texture
        rn = random.randrange (0, len(LTEXT))
        textureName=LTEXT[rn]
        info_msg ("chosen random = "+textureName)
        
    return(textureName)

def load_patterns(names_filter):    # load all patterns
    
    np = 0  # nr of patterns loaded
    LPATT = []  # list for mixed patterns
    (np, LPATT) = pdb.gimp_patterns_get_list (names_filter)       
    
    tempImage = pdb.gimp_image_new(256, 256, RGB)
    tempLayer = pdb.gimp_layer_new(tempImage,256,256,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
    tempImage.add_layer(tempLayer, 0)

    LLUMA = []  # list for LUMA value
    LCOLR = []  # list for PREVAILING COLOURS 
    
    if (np == 0):
        erro_msg("ERROR! Folder for patterns does not contain any valid pattern. Exit.")
        exit
    else :
        #info_msg ("info: nr of patterns = "+str(np))
        for i in range (np) :
            if LPATT[i] <> "":  # don't load null name patterns (fived)
                pdb.gimp_context_set_pattern (LPATT[i])
                pdb.gimp_drawable_fill (tempLayer, PATTERN_FILL)
    
                (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 1, 0, 255)
                (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 2, 0, 255)
                (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 3, 0, 255)
                aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))       #luminance=(red*0.299+green*0.587+blue*0.114)
                LLUMA.append(aCformula)
            
                meanCol = find_predominant_color (meanR, meanG, meanB)
                #info_msg("the predominant colour of this filler is "+meanCol)
                LCOLR.append(meanCol)        
            
    il=len(LCOLR)
    #for i in range(0, il):
        #info_msg(LCOLR[i])
    return(LPATT,LLUMA,LCOLR)

def load_gradients(names_filter):    # load all gradients
    
    ng = 0  # nr of gradients loaded
    LGRAD = []  # list for mixed gradients
    (ng, LGRAD) = pdb.gimp_gradients_get_list (names_filter)       # "Mandel"
    
    tempImage = pdb.gimp_image_new(256, 1, RGB)
    tempLayer = pdb.gimp_layer_new(tempImage,256,1,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
    tempImage.add_layer(tempLayer, 0)

    LLUMA = []  # list for grad formula value
    LCOLR = []  # list for PREVAILING COLOURS 
    
    if (ng == 0):
        erro_msg ("ERROR! Folder for gradients does not contain any valid gradient. Exit.")
        exit
    else :
        #info_msg ("info: nr of gradients = "+str(ng))
        for i in range (ng) :
        
            pdb.gimp_context_set_gradient (LGRAD[i])
            pdb.gimp_edit_blend (tempLayer, CUSTOM_MODE, NORMAL_MODE, GRADIENT_LINEAR, 100.0,
                             0, REPEAT_NONE, False, False, 1, 0, False, 0, 0, 255, 0)
    
            (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 1, 0, 255)
            (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 2, 0, 255)
            (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 3, 0, 255)
            aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))       #luminance=(red*0.299+green*0.587+blue*0.114)
            LLUMA.append(aCformula)
            
            meanCol = find_predominant_color (meanR, meanG, meanB)
            #info_msg("the predominant colour of this filler is "+meanCol)
            LCOLR.append(meanCol)        
            
    
    return(LGRAD,LLUMA,LCOLR)

############################################################################
#
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
############################################################################
#

# ************ # 
# MAIN PROGRAM #
# ------------ #

def TexturizeByAreaByColourSE(inImage, drawable,
                widthheight, 
                nr_of_colours,
                prep_whitebal, prep_stretchRGB,
                prep_curves,
                postp_artistic,
                postp_colors,
                namesFilter,
                fillType,
                selMethod,
                separation,
                BGonTop,
                doFlatten):
  
   
    # image characteristics
    width = inImage.width
    height = inImage.height
    type = inImage.base_type
    name = drawable.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    suffix = ""    
    
    # calculate scale factor for the new output image
    realWidthHeight = width+height      # real size of the input image  # bug fix 1.1.1
    if widthheight == 0 :               # keep original dimensions
        widthheight = realWidthHeight                                   # bug fix 1.1.1
        scale = 1
    else :    
        optiWidthHeight = widthheight       # size to work on to get best results
        scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling
        
    # check filter for names
    if namesFilter == "None" :
        namesFilter = ""    # no screening

# save current context
    pdb.gimp_context_push
    
# fix all required context settings 
    pdb.gimp_context_set_antialias(True)
    pdb.gimp_context_set_feather(False)
    pdb.gimp_context_set_feather_radius(0, 0)
    pdb.gimp_context_set_sample_merged(False) 
    pdb.gimp_context_set_sample_criterion(SELECT_CRITERION_COMPOSITE) 
    pdb.gimp_context_set_sample_threshold(0.00)
    pdb.gimp_context_set_sample_transparent(False)       # beware, if TRUE may change the area selected!
    pdb.gimp_context_set_interpolation(INTERPOLATION_LANCZOS)
    
#### START OPERATIONS +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    newwidth=int(width*scale)
    newheight=int(height*scale)
    #info_msg ("newwidth="+str(newwidth))
    newImage = pdb.gimp_image_new(newwidth, newheight, type)                    
    newBG = pdb.gimp_layer_new_from_drawable(drawable, newImage)     # background layer
    newBG.mode = NORMAL_MODE
    newBG.name = "Background Image Layer"
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    #pdb.gimp_desaturate_full(newBG, 1)
    pdb.gimp_context_set_interpolation(INTERPOLATION_LANCZOS)
    pdb.gimp_layer_scale(newBG, newwidth, newheight, False)
 
    alpha=pdb.gimp_drawable_has_alpha(newBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (newBG)    #  RGBA
        
    #pdb.gimp_display_new(newImage)    
    
# optional preprocessing
# ======================
    namesuffix = ""
    if prep_whitebal == True :
        # stretching the colours 
        newBG = pdb.python_fu_WhiteBalanceStretch (newImage, newBG)
        namesuffix = namesuffix+", white balanced"
    if prep_stretchRGB == True :
        # stretching the colours 
        newBG = pdb.python_fu_Stretch_RGB_colours (newImage, newBG, 1, 1)
        namesuffix = namesuffix+", RGB stretched"
        
    if prep_curves == 5 :   # auto adjust
        (mean, stddev, median, pixels, count, percent)=pdb.gimp_histogram(newBG, 0, 0, 255)
    # goal is a bit dark image, place centre at 96 instead of 128
        if mean > 144.0 :
            prep_curves = 4    # darken
        elif mean > 112.0 :    
            prep_curves = 3    # moderate darken
        elif mean > 80.0 :    
            prep_curves = 0    # none    
        elif mean > 48.0 :    
            prep_curves = 1    # moderate brighten
        else :    
            prep_curves = 2    # brighten
    if prep_curves == 1 :   # moderate brighten
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                ( 0, 0, 127, 159, 255, 255) 
                               )
        namesuffix = namesuffix+", brightened"
    if prep_curves == 2 :   # brighten
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                ( 0, 0, 127, 191, 255, 255) 
                               )
        namesuffix = namesuffix+", brightened"
    if prep_curves == 3 :   # moderate darken
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                ( 0, 0, 127, 95, 255, 255) 
                               )
        namesuffix = namesuffix+", darkened"
    if prep_curves == 4 :   # darken
        pdb.gimp_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                ( 0, 0, 127, 63, 255, 255) 
                               )
        namesuffix = namesuffix+", darkened"
        
# smooth
# ======
    smooth_layer=pdb.gimp_layer_new_from_drawable(newBG, newImage)
    smooth_layer.name="Smoothing layer"
    newImage.add_layer(smooth_layer, 0)
    pdb.gimp_layer_set_opacity(smooth_layer, 100.0)
    pdb.gimp_layer_set_mode(smooth_layer, SOFTLIGHT_MODE)
    pdb.plug_in_sel_gauss (newImage, smooth_layer, 10.0, 40)          # remove "noise" but...
    pdb.plug_in_unsharp_mask (newImage, smooth_layer, 5.0, 1.0, 0)    # increase contrast on details
    smooth_layer=newImage.active_layer

# create quantized layer
# ======================

    quantized_layer=pdb.gimp_layer_new_from_drawable(smooth_layer, newImage) 
    pdb.gimp_item_set_name(quantized_layer, "Quantized Layer "+str(nr_of_colours))  
    newImage.add_layer(quantized_layer, 0)
    pdb.gimp_layer_set_opacity(quantized_layer, 100.0)
    pdb.gimp_layer_set_mode(quantized_layer,NORMAL_MODE)

# quantize the colour_areas using mode indexed
# ============================================

# change image mode to indexed (this produces a cartoonized image, 4...64 colors)
    
    pdb.gimp_image_convert_indexed(newImage, 0, 0, nr_of_colours, 0, 0, "ignored")
    is_indexed=pdb.gimp_drawable_is_indexed(quantized_layer) 
    if is_indexed==1:
        hidden_message="OK, converted"
    else:    
        erro_msg ("ERROR! image not converted, exit")
        return    

# get the number of pixels in the colormap (which exists for indexed images)
# ==========================================================================
    (count, colors)=pdb.gimp_image_get_colormap(newImage)
    # the count is nr_of_colors * 3, example 8*3=24
    nr_areas=(count/3) 
    ##info_msg("nr of colors in the palette: "+str(nr_areas)) 
  
    temp_palette=pdb.gimp_palette_new(pdb.gimp_image_get_name(newImage))
    
    index=0
    while index<count:
# Compute the color
        color= (colors[index], colors[index+1], colors[index+2])
# Add to the filter temporary palette
        pdb.gimp_palette_add_entry(temp_palette, str(index/3), color)
        index+=3 
# Endwhile index<count
  
# the temporary palette has been created            
# ======================================

# re-set image to RGB to proceed
# ==============================
    pdb.gimp_image_convert_rgb(newImage)
    alpha=pdb.gimp_drawable_has_alpha(quantized_layer) 
    if alpha ==0:
        pdb.gimp_layer_add_alpha(quantized_layer)     #  RGBA

    (mean, stddev, median, pixels, count, percent)=pdb.gimp_histogram(quantized_layer, 0, 0, 255)
    avgpixels = pixels / nr_areas   # average number of pixels per colour area
    bigareapixels = pixels / 4      # define only a few areas as big, likely just one
    midareapixels = pixels / 64     # define some medium colour areas
    smallareapixels = pixels / 256   # define some small colour areas

#   loop to fill the N areas for colors with wood patterns            <<<<<<<<<<< for all N areas >>>>>>>>>>>
#   ======================================================

    pdb.gimp_image_undo_freeze(newImage)                      # save time and space
    
    if fillType == 0 :   # patterns
        (LTEXT, LLUMA, LCOLR)=load_patterns(namesFilter)
    if fillType == 1 :   # gradients
        (LTEXT, LLUMA, LCOLR)=load_gradients(namesFilter)
    
    #pdb.gimp_display_new(newImage)
    #return
    
    if separation == 2 :    # needlepoints ("embroidery")
        pdb.gimp_selection_none(newImage)
        embroidery = pdb.gimp_layer_new_from_drawable (quantized_layer, newImage)   # untextured
        newImage.add_layer (embroidery, 0)
        pdb.plug_in_spread (newImage, embroidery, 14, 14)
        pdb.plug_in_neon (newImage, embroidery, 7, 0.28)
    
    areaNR=0
    area_color=[]   # empty list
    for areaNR in range(0,nr_areas):
        one_color=pdb.gimp_palette_entry_get_color(temp_palette, areaNR)
        #info_msg ("one_color = "+str(one_color))
        if selMethod == 0 : # best fit in luma
            best_texture=choose_appropriate_texture_by_LUMA(one_color, LLUMA, LTEXT)
        elif selMethod == 1 : # best fit in colour
            best_texture=choose_appropriate_texture_by_COLOUR(one_color, LCOLR, LTEXT)
        elif selMethod == 2 : # best fit in colour&luma    
            best_texture=choose_appropriate_texture_by_LUMACOLOUR(one_color, LCOLR, LLUMA, LTEXT)
        else :    # use random texture (no option for now)
            rn = random.randrange (0, len(LTEXT))
            best_texture = LTEXT[rn]     
        pdb.gimp_image_select_color(newImage, CHANNEL_OP_REPLACE, quantized_layer, one_color)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(newImage)
        if there_is_a_selection==1:
            (mean, stddev, median, pixels, count, percent)=pdb.gimp_histogram(quantized_layer, 0, 0, 255)
            #info_msg ("pixels = "+str(pixels))
            
            azimuth_value = 135.0
            if pixels >= bigareapixels :
                width_value = widthheight/300.0  
                filter_mode = 1 # carve 
            elif pixels >= midareapixels :
                width_value = widthheight/600.0
                filter_mode = 0 # chisel
            elif pixels >= smallareapixels :
                width_value = widthheight/1200.0
                filter_mode = 0 # chisel
            else :  # very snall areas
                width_value = widthheight/2400.0
                filter_mode = 0 # chisel
            elev_value = 30     # after tests it looks as the best setting
            depth_value = 20    # after tests it looks as the best setting  
            #info_msg ("area nr. "+str(areaNR)+" there_is_a_selection")
            color_area=pdb.gimp_image_get_selection(newImage)
            this_channel=pdb.gimp_selection_save(newImage)
            
            if fillType == 0 :   # patterns
                pdb.gimp_context_set_pattern (best_texture)
                pdb.gimp_edit_fill (quantized_layer, PATTERN_FILL)
            if fillType == 1 :   # gradients
                pdb.gimp_context_set_gradient (best_texture)
                (flag, x1, y1, x2, y2) = pdb.gimp_selection_bounds (newImage)
                centrex = (x1+x2)/2
                centrey = (y1+y2)/2
                gtype = random.randrange (1,10) # omit simple linear, not good when start from centre
                pdb.gimp_edit_blend (quantized_layer, CUSTOM_MODE, NORMAL_MODE, gtype, 100.0,
                             0, REPEAT_NONE, False, False, 1, 0, False, centrex, centrey, x2, y2)
                pdb.plug_in_gauss_iir2 (newImage, quantized_layer, newImage.width/200, newImage.height/200)
                
            #try
            
            if separation < 2 :
                if separation == 0 :    # hard chisel/carve
                    ccRoundness = 0.1
                else :                  # soft chisel/carve
                    ccRoundness = 1.0
                #info_msg ("pixels="+str(pixels)+",smallareapixels="+str(smallareapixels))    
                if pixels >= smallareapixels :  # no chisel for very small areas
                    #info_msg ("chisel-width="+str(width_value))
                    #info_msg ("widthheight="+str(widthheight))
                    pdb.script_fu_chisel(newImage, quantized_layer,
                             width_value,             #bevel width
                             10.0*ccRoundness,        #bevel softness,
                             ccRoundness,             #bevel roundness (0.1-1.0),
                             ccRoundness,             #bevel power, #test
                             azimuth_value,           #azimuth  
                             elev_value,              #elevation
                             depth_value,             #depth,  
                             filter_mode,             #mode: 0=chisel,1=carve,
                             0,                       #location,
                             0.0,                     #post-effect blur,
                             False)                   #keep)
            
            #pdb.gimp_display_new(newImage)
            #return
            
                    LA = newImage.layers
            #il = len(LA)
            #for i in range(0, il):          
            #   info_msg ("layer "+str(i)+" has name "+LA[i].name)

                    bevel=LA[0]    # should be: ....bevel
                    quantized_layer=pdb.gimp_image_merge_down(newImage, bevel, 2)  # should merge ...bevel into quantized
                #else :
                    #info_msg ("no chisel on color "+str(one_color))
                pdb.gimp_item_set_name(quantized_layer, name+" (colours="+str(int(nr_of_colours))+")")

            #pdb.gimp_display_new(newImage)
            #return
        
           
    ##info_msg ("end loop")
    
    if separation == 2 :
        pdb.gimp_selection_none (newImage)
        embroidery.mode = LIGHTEN_ONLY_MODE
        quantized_layer=pdb.gimp_image_merge_down (newImage, embroidery, 2)
        
        #pdb.gimp_display_new (newImage)
        #return
        
    pdb.gimp_selection_none(newImage)
    pdb.gimp_layer_set_opacity(quantized_layer, 100.0)
    pdb.gimp_layer_set_mode(quantized_layer, GRAIN_MERGE_MODE)
    
    visible_layer=pdb.gimp_layer_new_from_visible(newImage, newImage, "Now visible Texture: "+namesFilter)
    newImage.add_layer(visible_layer, 0)
    visible_layer.opacity = 100.0
    pdb.gimp_desaturate_full (visible_layer, DESATURATE_LUMINOSITY)
    pdb.plug_in_vinvert (newImage, visible_layer)   
    pdb.gimp_layer_set_mode (visible_layer,SOFTLIGHT_MODE)
    #pdb.gimp_item_set_visible (smooth_layer, False) # not to have a double effect
    pdb.gimp_item_set_visible (newBG, False) # not to have a double effect
        
    #pdb.gimp_display_new (newImage)
    #return
        
    # Post-processing: Artistic
    if postp_artistic>0:
        artistic_layer=pdb.gimp_layer_new_from_visible( newImage, newImage, "Artistic Layer")
        newImage.add_layer(artistic_layer, 0)
        
        #pdb.plug_in_vinvert (newImage, artistic_layer)
        if postp_artistic==1:  # graphic boost (increase contrast)
            pdb.gimp_brightness_contrast (artistic_layer, +25, +50)
            artistic_layer.name="Contrast Boost"
            pdb.gimp_layer_set_mode(artistic_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 60.0)
        if postp_artistic==2:  # graphic evidence
            pdb.plug_in_erode (newImage, artistic_layer, 1, HISTOGRAM_VALUE, 1.0, 7, 0, 128)
            pdb.plug_in_erode (newImage, artistic_layer, 1, HISTOGRAM_VALUE, 1.0, 7, 0, 128)
            pdb.gimp_color_balance (artistic_layer, 1, True, -50, +20, +20)    # 1=MIDTONES
            artistic_layer.name="Graphic Evidence"
            pdb.gimp_layer_set_mode(artistic_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 60.0)
        if postp_artistic==3:  # anguish
            pdb.script_fu_difference_clouds(newImage, artistic_layer)
            artistic_layer = pdb.gimp_image_get_active_layer (newImage)
            artistic_layer.name="Anguish"
            pdb.gimp_layer_set_mode(artistic_layer, COLOR_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 75.0)
        if postp_artistic==4:  # colored pencils
            Preset = "pencildraw"
            pdb.plug_in_gimpressionist (newImage, artistic_layer, Preset)
            artistic_layer.name="Colored Pencils"
            pdb.gimp_layer_set_mode(artistic_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 45.0)
        if postp_artistic==5:  # make squiggly
            Preset = "squiggly"
            pdb.plug_in_gimpressionist (newImage, artistic_layer, Preset)
            artistic_layer.name="Squiggly"
            pdb.gimp_layer_set_mode(artistic_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 45.0)
        if postp_artistic==6:  # edges
            pdb.plug_in_edge (newImage, artistic_layer, 1.0, 1, 4)    # 4 = DIFFERENTIAL
            artistic_layer.name="Edges"
            pdb.plug_in_vinvert (newImage, artistic_layer)
            pdb.plug_in_colortoalpha (newImage, artistic_layer, ((255,255,255,255)))
            pdb.gimp_layer_set_mode(artistic_layer, HARDLIGHT_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 75.0)
        if postp_artistic==7:  # noise
            Preset = "dither"
            pdb.plug_in_gimpressionist (newImage, artistic_layer, Preset)
            artistic_layer.name="Dither"
            pdb.gimp_layer_set_mode(artistic_layer, DISSOLVE_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 60.0)
        if postp_artistic==8:  # colored texture 
            pdb.gimp_context_set_pattern ("Leather-05.jpg")     
            pdb.gimp_edit_fill (artistic_layer, PATTERN_FILL) 
            artistic_layer.name="Colored Texture"
            pdb.gimp_layer_set_mode(artistic_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(artistic_layer, 35.0)
    #pdb.gimp_display_new(newImage)
    #return
    # Post-processing: Colors
    if postp_colors>0:
        colors_layer=pdb.gimp_layer_new_from_visible(newImage, newImage, "Colors layer")
        newImage.add_layer(colors_layer, 0)
        
        if postp_colors==1:  # chroma boost
            pdb.gimp_hue_saturation (colors_layer, 0, 0, 0, +66)    
            colors_layer.name="Chroma Boost"
            pdb.gimp_layer_set_mode(colors_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 75.0)
        if postp_colors==2:  # color balance (only if not executed as initial step)
            if prep_whitebal == False :
                newBG = pdb.python_fu_WhiteBalanceStretch (newImage, colors_layer)
            colors_layer.name="Balance Gamma"
            pdb.gimp_layer_set_mode(colors_layer, NORMAL_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 65.0)
        if postp_colors==3:  # sepia
            # Sepia: RGB=56,44,30
            pdb.gimp_context_set_foreground ((45,44,30,255))
            pdb.gimp_edit_fill (colors_layer, FOREGROUND_FILL)
            colors_layer.name="Sepia"
            pdb.gimp_layer_set_mode(colors_layer, COLOR_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 75.0)
        if postp_colors==4:  # metallic look - silver
            # Silver: RGB=138,149,151
            pdb.gimp_context_set_foreground ((138,149,151,255))
            pdb.gimp_edit_fill (colors_layer, FOREGROUND_FILL)
            colors_layer.name="Metallic - Silver"
            pdb.gimp_layer_set_mode(colors_layer, COLOR_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 90.0)
        if postp_colors==5:  # metallic look - copper
            # Copper: RGB=142,064,042
            pdb.gimp_context_set_foreground ((142,064,042,255))
            pdb.gimp_edit_fill (colors_layer, FOREGROUND_FILL)
            colors_layer.name="Metallic - Copper"
            pdb.gimp_layer_set_mode(colors_layer, COLOR_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 80.0)
        if postp_colors==6:  # metallic look - blue steel
            # Blue Steel: RGB=035,026,036
            pdb.gimp_context_set_foreground ((035,026,036,255))
            pdb.gimp_edit_fill (colors_layer, FOREGROUND_FILL)
            colors_layer.name="Metallic - BlueSteel"
            pdb.gimp_layer_set_mode(colors_layer, COLOR_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 95.0)
        if postp_colors==7:  # textured glass
            pdb.plug_in_glasstile (newImage, colors_layer, 10, 10)    
            colors_layer.name="Textured Glass"
            pdb.gimp_layer_set_mode(colors_layer, LIGHTEN_ONLY_MODE)
            pdb.gimp_layer_set_opacity(colors_layer, 75.0)
    
    if BGonTop == True :
        pdb.gimp_image_raise_item_to_top (newImage, newBG)
        pdb.gimp_item_set_visible (newBG, True) # restore visibility
        pdb.gimp_layer_set_mode (newBG, NORMAL_MODE)
        newBG.opacity = 35.0 
        pdb.gimp_image_set_active_layer (newImage, newBG)       # fix 1.1.1
        dupBG=pdb.gimp_layer_new_from_drawable(newBG, newImage)
        dupBG.name=newBG.name+" duplicated "
        newImage.add_layer(dupBG, 0)
        WhatYouSee = pdb.gimp_layer_new_from_visible(newImage, newImage, "What You See")
        newImage.add_layer(WhatYouSee, 0)
       # decide which Mode and Opacity depending on the MeanValue of ex-BG layer     
        (meanValue, stdDev, median, pix, cnt, perc) = pdb.gimp_histogram (WhatYouSee, HISTOGRAM_VALUE, 0 ,255)
       #info_msg ("meanValue = "+str(meanValue))
        if meanValue < 64 :    # dark  
            pdb.gimp_layer_set_mode (WhatYouSee, DODGE_MODE)
            pdb.gimp_layer_set_opacity (WhatYouSee, 75.0)
        elif meanValue > 192 :    # light  
            pdb.gimp_layer_set_mode (WhatYouSee, MULTIPLY_MODE)
            pdb.gimp_layer_set_opacity (WhatYouSee, 65.0)
        else :  # average      
            pdb.gimp_layer_set_mode (WhatYouSee, SOFTLIGHT_MODE)
            pdb.gimp_layer_set_opacity (WhatYouSee, 55.0)
            
    # if both Post-processings = none, you may set visible only the texture layer by default     
    # if postp_artistic==0 and postp_colors==0 and BGonTop == False:
    #    pdb.gimp_item_set_visible (visible_layer, False)

#   finish the script

    pdb.gimp_palette_delete(temp_palette)
    
    pdb.gimp_context_pop
    
    if (doFlatten == True) :
        pdb.gimp_image_flatten(newImage)
        flattenedLayer = pdb.gimp_image_get_active_layer (newImage)
        pdb.gimp_item_set_name(flattenedLayer,
            name+namesuffix+", Texturized (colours="+str(int(nr_of_colours))+"; texture="+namesFilter+")")
  
    pdb.gimp_display_new(newImage)

    return

# This is the plugin registration function  
register(
    "TexturizeByAreaByColourSE",
    "Create a quantized image with chosen texture effects (special edition: no GMIC)",
    "Transform an RGB image into an indexed image and apply a chosen texture effect on each colour area",
    "Diego", 
    "Diego Nassetti ", 
    "November 2016",
    "TexturizeByAreaByColourSE (v.1.2 no-GMIC)",
    "RGB*",
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "drawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image (0=same as source)", 4800, (0, 9600, 300)),
      (PF_ADJUSTMENT, "nuances", "How many Nuances", 32, (4, 256, 1)),
      (PF_TOGGLE, "whitebal", "WhiteBalance preprocess?", True),
      (PF_TOGGLE, "rgbstretch", "Stretch RGB preprocess?", True),
      (PF_OPTION, "curve", "Curves preprocess?",  5, ("None",
                                                         "Moderate Brighten",
                                                         "Brighten",
                                                         "Moderate Darken",
                                                         "Darken",
                                                         "Auto Adjust",
                                                        )),
      (PF_OPTION, "postpart", "Artistic PostProcess", 0, ("None",                                                          
                                                          "Contrast Boost",
                                                          "Graphic Evidence",
                                                          "Anguish",
                                                          "Colored Pencil",
                                                          "Squiggly",
                                                          "Edges",
                                                          "Dithering",
                                                          "Colored Texture",
                                                          )),
      (PF_OPTION, "postpcol", "Colors PostProcess", 0, ("None",
                                                        "Chroma Boost",
                                                        "Color Balance",
                                                        "Sepia",
                                                        "Metallic Look - Silver",
                                                        "Metallic Look - Copper",
                                                        "Metallic Look - BlueSteel",
                                                        "Textured Glass",
                                                        )),
      (PF_STRING, "filter", "Filter for fillers names (None=use all):", "None"),
      (PF_OPTION, "filltype", "Type of fillers", 0, ("Patterns",
                                                     "Gradients",
                                                     )),
      (PF_OPTION, "selmethod", "Fillers selection method", 2, ("Best Luma",
                                                               "Best Colour",
                                                               "Best Luma&Colour",
                                                               )),
      (PF_OPTION, "separator", "Areas separation method", 1, ("Chisel/Carve, roundness 0.1",
                                                              "Chisel/Carve, roundness 1.0",
                                                              "Needlepoint", "None",
                                                               )),
      (PF_TOGGLE, "bgontop", "Move BG on top?", True),
      (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    TexturizeByAreaByColourSE,
    menu="<Image>/Diego/Quantize",
    )

main()
